#!/usr/bin/env python3
'''
Merge experiments into a common directory.
'''

import argparse
import logging
import pathlib
import shutil

import yaml

from mlflow_extra.common import configure_logging
from mlflow_extra.metadata import Metadata

LOGGER = logging.getLogger(__name__)


def update_experiment_ids(path, new_id):
    '''
    Update the experiment ID in all metadata files under the given path.

    Args:
        path:
            The path under which to update the experiment IDs.

        new_id:
            The new experiment ID.
    '''
    for data in Metadata().find_meta(path):
        old_id = data.experiment_id
        if old_id != new_id:
            LOGGER.info(
                'Replacing experiment ID %s with %s in %s',
                old_id,
                new_id,
                data.path
            )
            data.experiment_id = new_id
            data.save()


def fix_experiment_ids(path):
    '''
    Fix the experiment IDs for the experiment and all runs under the given path.
    This will ensure that the experiment ID is a non-negative integer and that
    the ID matches the experiment's directory name. The directory will be moved
    if necessary.

    Args:
        path:
            The path to the experiment directory.
    '''
    path = pathlib.Path(path).resolve()

    # Experiment ID according to current directory path.
    try:
        dir_id = Metadata.non_negative_int(path.name)
    except ValueError:
        pass
    else:
        update_experiment_ids(path, dir_id)
        return

    # Invalid experiment ID in directory name.
    metadata = Metadata(path=path)
    try:
        exp_id = metadata.non_negative_int(metadata.experiment_id)
    except ValueError:
        pass
    else:
        # Check if the directory can be renamed to the current ID.
        new_path = path.parent / str(exp_id)
        if not new_path.exists():
            shutil.move(path, new_path)
            update_experiment_ids(new_path, exp_id)
            return

    # Find the first available directory path to set a new experiment ID.
    exp_id = 0
    while True:
        new_path = path.parent / str(exp_id)
        if new_path.exists():
            exp_id += 1
            continue
        shutil.move(path, new_path)
        update_experiment_ids(new_path, exp_id)
        return


def reorder_experiment_ids(path):
    '''
    Reorder all of the experiment IDs in a directory by their creation time.
    '''
    path = pathlib.Path(path).resolve()
    metadatas = (
        Metadata(path=mpath)
        for mpath in path.glob(f'*/{Metadata.FILENAME}')
    )
    metadatas = sorted(metadatas, key=lambda dat: (dat.creation_time, dat.name))
    prefix = 'tmp_'
    i = 0
    while list(path.glob(f'{prefix}*')):
        prefix = f'tmp_{i:d}_'
        i += 1
    for i, metadata in enumerate(metadatas):
        tmp_path = path / f'{prefix}{i:d}'
        LOGGER.info('moving %s to temporary path %s', metadata.path.parent, tmp_path)
        shutil.move(metadata.path.parent, tmp_path)
    for tmp_path in path.glob(f'{prefix}*'):
        new_path = tmp_path.parent / tmp_path.name[len(prefix):]
        LOGGER.info('moving temporary path %s to %s', tmp_path, new_path)
        shutil.move(tmp_path, new_path)
        fix_experiment_ids(new_path)


def collect_experiment_names(path):
    '''
    Collect all experiment names from the experiment directories under the given
    path.

    Args:
        path:
            The directory path containing the experiment directories.

    Returns:
        A dict mapping experiment names to lists of experiment directory paths.
    '''
    experiments = {}
    path = pathlib.Path(path).resolve()
    for metadata_path in path.glob(f'*/{Metadata.FILENAME}'):
        metadata = Metadata(path=metadata_path)
        experiments.setdefault(metadata.name, []).append(metadata)
    return experiments


def merge_experiments_by_name(path):
    '''
    Merge runs from experiments with the same name into a single experiment.

    Args:
        path:
            The directory path containing the experiment directories.
    '''
    experiments = collect_experiment_names(path)
    for name, metadatas in experiments.items():
        if len(metadatas) < 2:
            continue
        target_md = metadatas[0]
        for metadata in metadatas[1:]:
            exp_dir = metadata.path.parent
            if target_md.lifecycle_stage != metadata.lifecycle_stage:
                LOGGER.warning(
                    (
                        'Refusing to merge experiments named %s in different lifecycle stages:'
                        '%s [%s] and %s [%s]'
                    ),
                    name,
                    target_md.path, target_md.lifecycle_stage,
                    metadata.path, metadata.lifecycle_stage
                )
                continue
            target_md.creation_time = min(target_md.creation_time, metadata.creation_time)
            target_md.last_update_time = min(target_md.last_update_time, metadata.last_update_time)
            for run_dir in exp_dir.glob('*'):
                if not run_dir.is_dir():
                    continue
                new_run_dir = target_md.path.parent / run_dir.name
                if new_run_dir.exists():
                    LOGGER.warning(
                        'Run %s already exists in %s: %s will not be merged',
                        run_dir.name, new_run_dir.parent, run_dir
                    )
                    continue
                LOGGER.info('Moving run from %s to %s', run_dir, new_run_dir)
                shutil.move(run_dir, new_run_dir)
            shutil.rmtree(exp_dir)
        update_experiment_ids(target_md.path.parent, target_md.experiment_id)


def copy_experiments(src_path, dst_path):
    '''
    Copy experiments from one path to another.

    Args:
        src_path:
            The source directory from which to move the experiments.

        dst_path:
            The target directory to which to move the experiments.
    '''
    src_path = pathlib.Path(src_path).resolve()
    dst_path = pathlib.Path(dst_path).resolve()

    next_exp_id = 0
    for path in dst_path.glob('*'):
        try:
            exp_id = int(path.name)
        except ValueError:
            continue
        next_exp_id = max(next_exp_id, exp_id)
    next_exp_id += 1

    for path in src_path.glob('*'):
        if not path.is_dir() or path.name == '.trash':
            continue
        new_path = dst_path / path.name
        while new_path.exists():
            new_path = dst_path / str(next_exp_id)
            next_exp_id += 1
        LOGGER.info('Copying %s to %s', path, new_path)
        shutil.copytree(path, new_path)


def fix_artifact_uris(path, map_path=None):
    '''
    Update artifact URIs for all experiments and runs under the given path.

    Args:
        path:
            The path containing the experiments to fix.

        map_path:
            An optional path to a YAML file that maps old paths to new ones.
    '''
    path_map = {}
    if map_path is not None:
        with open(map_path, 'rb') as handle:
            data = yaml.safe_load(handle)
        for old, new in data.items():
            old = pathlib.Path(old).resolve()
            new = pathlib.Path(new).resolve()
            path_map[old] = new

    for data in Metadata().find_meta(path):
        uri = data.artifact_uri
        # Skip anything that isn't a local path.
        if not uri.is_local_file:
            continue

        current_path = uri.path
        new_path = None

        # Update from the path map.
        for old, new in path_map.items():
            if current_path.is_relative_to(old):
                subpath = current_path.relative_to(old)
                new_path = new / subpath
                break

        else:
            # Assume that an artifact directory in the same directory is the
            # intended target.
            artifact_dir = data.path.parent / 'artifacts'
            if artifact_dir.exists():
                new_path = artifact_dir

            # Assume that experiments point to their own directory be default.
            elif data.is_experiment and not uri.path.exists():
                new_path = data.path.parent

        if new_path is not None and new_path != current_path:
            LOGGER.info('Updating artifact URI to %s in %s', new_path, data.path)
            uri.path = new_path
            data.artifact_uri = uri
            data.save()
            continue

        # Warn about missing paths.
        if not current_path.exists():
            LOGGER.warning(
                'Artifacts URI does not exist: %s [%s]',
                uri,
                data.path
            )


def script_merge_experiments(args=None):
    '''
    Copy experiments into a common MLflow directory. Runs from experiments with
    the same name will be merged.
    '''
    configure_logging()

    parser = argparse.ArgumentParser(description=script_merge_experiments.__doc__)
    parser.add_argument(
        'target',
        help='The directory into which to merge the experiments. Default: %(default)s'
    )
    parser.add_argument(
        'dirs',
        nargs='+',
        help='The directories with the experiments to merge.'
    )
    pargs = parser.parse_args(args=args)

    target_path = pathlib.Path(pargs.target).resolve()
    target_path.mkdir(parents=True, exist_ok=True)

    for path in pargs.dirs:
        copy_experiments(path, target_path)
    merge_experiments_by_name(target_path)
    reorder_experiment_ids(target_path)
    fix_artifact_uris(target_path)


def script_fix_artifact_uris(args=None):
    '''
    Attempt to fix broken artifact URIs in experiments and runs.
    '''
    configure_logging()

    parser = argparse.ArgumentParser(description=script_fix_artifact_uris.__doc__)
    parser.add_argument(
        'path',
        help='A path to a directory with experiments and runs.'
    )
    parser.add_argument(
        '-m', '--map',
        help='A path to a YAML file that maps old paths to new paths.'
    )
    pargs = parser.parse_args(args=args)
    fix_artifact_uris(pargs.path, map_path=pargs.map)


def script_fix_experiment_ids(args=None):
    '''
    Attempt to fix experiment IDs so that the experiment's directory and all of
    its runs match its ID.
    '''
    configure_logging()

    parser = argparse.ArgumentParser(description=script_fix_experiment_ids.__doc__)
    parser.add_argument(
        'paths',
        nargs='+',
        help='Experiment directory paths.'
    )
    pargs = parser.parse_args(args=args)
    for path in pargs.paths:
        try:
            fix_experiment_ids(path)
        except FileNotFoundError as err:
            LOGGER.error('failed to fix IDs in %s: %s', path, err)


if __name__ == '__main__':
    script_merge_experiments()



#!/usr/bin/env python3
'''
URI manipulation.
'''

import logging
import pathlib
import urllib.parse


LOGGER = logging.getLogger(__name__)
FILENAME = 'meta.yaml'


class URITransformer():
    '''
    Transform URIs, including file paths.
    '''
    def __init__(self, uri):
        self.uri = uri
        self.parts = list(urllib.parse.urlparse(uri))
        self.path = self.parts[2]

    def __str__(self):
        parts = self.parts.copy()
        parts[2] = str(parts[2])
        if not parts[0]:
            return parts[2]
        return urllib.parse.urlunparse(parts)

    @property
    def scheme(self):
        '''
        The URI scheme.
        '''
        return self.parts[0]

    @property
    def is_local_file(self):
        '''
        True if the URI points to a local path, else False. The path may not exist.
        '''
        scheme = self.scheme
        return (not scheme) or scheme.lower() == 'file'

    def _get_path(self):
        return self.parts[2]

    def _set_path(self, path):
        path = pathlib.Path(path)
        if self.is_local_file and not path.is_absolute():
            path = path.resolve()
        self.parts[2] = path

    path = property(
        fget=_get_path,
        fset=_set_path,
        doc='The URI path component'
    )



#!/usr/bin/env python3
'''
Common functionality.
'''

import logging


LOGGER = logging.getLogger(__name__)
DEFAULT_DIRECTORY = 'mlruns'


def configure_logging(level=logging.INFO):
    '''
    Configure logging.

    Args:
        level:
            The logging level.
    '''
    logging.basicConfig(
        style='{',
        format='[{asctime}] {levelname} {message}',
        datefmt='%Y-%m-%d %H:%M:%S',
        level=level
    )



#!/usr/bin/env python3

'''
Delete suboptimal runs.
'''

import argparse
import logging
import pathlib
import subprocess
import tempfile

import pandas as pd

from mlflow.tracking import _get_store

from mlflow_extra.common import configure_logging


LOGGER = logging.getLogger(__name__)
METRIC_PREFIX = 'metrics.'


def export_csv(experiment_id, path):
    '''
    Export experiment data to a CSV file.

    Args:
        experiment_id:
            The experiment ID.

        path:
            The output path.
    '''
    path = pathlib.Path(path).resolve()
    path.parent.mkdir(parents=True, exist_ok=True)
    LOGGER.debug('Exporting data from experiment %d', experiment_id)
    subprocess.run(
        ('mlflow', 'experiments', 'csv', '-x', str(experiment_id), '-o', str(path)),
        stdout=subprocess.DEVNULL,
        check=True
    )


def list_metrics(data):
    '''
    Print metric data to STDOUT.

    Args:
        data:
            The data loaded from the CSV file.
    '''
    data = data.filter(regex=f'^{METRIC_PREFIX}', axis=1)
    data.columns = [col[len(METRIC_PREFIX):] for col in data.columns]
    data = data[sorted(data.columns)]
    for _metric, values in data.iteritems():
        print(values.describe())
        print()


def filter_by_threshold(data, thresholds, ascending=False):
    '''
    Filter rows in data by threshold values.

    Args:
        data:
            The data to filter.

        thresholds:
            A dict mapping column names to threshold values.


        ascending:
            If True, keep values less than or equal to the thresold values,
            otherwise keep values greater than or equal to the threshold values.
    '''
    for name, value in thresholds.items():
        if ascending:
            data = data[data[name] <= value]
        else:
            data = data[data[name] >= value]
    return data


def filter_by_number(data, columns, number, ascending=False):
    '''
    Filter rows to only keep a selected number.

    Args:
        data:
            The data to filter.

        columns:
            The columns by which to filter the data before selecting the rows.

        number:
            The number of rows to keep.

        ascending:
            If True, sort the data in ascending order before selecting the rows
            to keep, otherwise sort in descending order.
    '''
    return data.sort_values(by=columns, ascending=ascending).head(n=number)


def delete_runs(run_ids, total_runs, confirm=False):
    '''
    Delete the selected runs

    Args:
        run_ids:
            An iterable of run IDs to delete.

        total_runs:
            The total number of runs in the data, only used for display.

        confirm:
            If True, delete the run, else perform a dryrun.

    Returns:
        True if anything was deleted, False otherwise.
    '''
    LOGGER.debug('Loading MLflow data store')
    store = _get_store()
    n_runs = len(run_ids)
    LOGGER.info(
        '%d / %d run%s selected for deletion',
        n_runs,
        total_runs,
        '' if n_runs == 1 else 's'
    )
    prefix = '' if confirm else '[DRYRUN] '
    deleted = False
    for i, run_id in enumerate(run_ids, start=1):
        LOGGER.info('%sDeleting run %s [%d / %d]', prefix, run_id, i, n_runs)
        if confirm:
            #  subprocess.run(
            #      ('mlflow', 'runs', 'delete', '--run-id', str(run_id)),
            #      stdout=subprocess.DEVNULL,
            #      check=True
            #  )
            store.delete_run(run_id)
            deleted = True
    if deleted:
        print('Runs can be restored with mlflow runs restore --run-id <run_id>')

    return deleted


def script_filter_runs(args=None):  # pylint: disable=too-many-locals
    '''
    Delete runs from experiment based on thresholds.
    '''
    configure_logging()

    parser = argparse.ArgumentParser(description=script_filter_runs.__doc__)
    parser.add_argument(
        'experiment_id',
        type=int,
        nargs='+',
        help='The MLflow experiment ID (see mlflow experiments list).'
    )
    parser.add_argument(
        '-a', '--ascending',
        action='store_true',
        help='Keep the first n runs in ascending order instead of descending.'
    )
    parser.add_argument(
        '-c', '--confirm',
        action='store_true',
        help='Confirm the deletion. Without this only a dryrun is performed.'
    )
    parser.add_argument(
        '-l', '--list',
        action='store_true',
        help='List metrics and their statistics.'
    )
    parser.add_argument(
        '-m', '--metrics',
        nargs='+',
        action='extend',
        help='The metrics by which to filter runs.'
    )
    parser.add_argument(
        '-n', '--number',
        type=int,
        help='The number of runs to keep.'
    )
    parser.add_argument(
        '-t', '--thresholds',
        nargs='+',
        action='extend',
        type=float,
        help=('''
            The threshold values for the selected metrics. In descending order
            the threshold values are a lower limit. In ascending order they are
            an upper limit.
        ''')
    )
    pargs = parser.parse_args(args=args)
    has_req_args = pargs.metrics and (pargs.thresholds or pargs.number is not None)
    if not (has_req_args or pargs.list):
        LOGGER.warning('You must select either a number of a threshold value.')
        return

    run_ids = set()
    total_runs = 0
    with tempfile.TemporaryDirectory() as tmpdir:
        for experiment_id in pargs.experiment_id:
            # Extract the run data to a temporary CSV file.
            tmpdir = pathlib.Path(tmpdir)
            csv_path = tmpdir / 'run.csv'
            export_csv(experiment_id, csv_path)
            data = pd.read_csv(csv_path)
            if not data:
                LOGGER.warning('no data for experiment %s', experiment_id)
                continue

            # List data and exit if requested.
            if pargs.list:
                list_metrics(data)
                continue

            total_runs += data.shape[0]
            data_to_keep = data
            metrics = [METRIC_PREFIX + m for m in pargs.metrics]

            # Filter by selected metrics.
            if pargs.thresholds:
                data_to_keep = filter_by_threshold(
                    data_to_keep,
                    dict(zip(metrics, pargs.thresholds)),
                    ascending=pargs.ascending
                )

            # Filter to the selected number of runs.
            if isinstance(pargs.number, int) and pargs.number >= 0:
                data_to_keep = filter_by_number(
                    data_to_keep, metrics, pargs.number, ascending=pargs.ascending
                )
            run_ids.update(set(data['run_id']) - set(data_to_keep['run_id']))

    delete_runs(run_ids, total_runs, confirm=pargs.confirm)


if __name__ == '__main__':
    script_filter_runs()



#!/usr/bin/env python3
'''
Find and manipulate MLflow metadata files.
'''

import logging
import os
import pathlib
import shutil
import tempfile

import yaml

from mlflow_extra.uri import URITransformer

LOGGER = logging.getLogger(__name__)


class Metadata():
    '''
    Metadata wrapper.
    '''
    FILENAME = 'meta.yaml'

    FIELD_ARTIFACT_LOCATION = 'artifact_location'  # experiment metadata
    FIELD_ARTIFACT_URI = 'artifact_uri'  # run metadata
    FIELD_CREATION_TIME = 'creation_time'
    FIELD_EXPERIMENT_ID = 'experiment_id'
    FIELD_LAST_UPDATE_TIME = 'last_update_time'
    FIELD_LIFECYCLE_STAGE = 'lifecycle_stage'
    FIELD_NAME = 'name'

    def __init__(self, path=None, data=None):
        '''
        Args:
            path:
                The path to a metadata file, or a directory that contains one.
                If not None, then the data will be loaded from the path unless
                data is also passed via the data parameter.

            data:
                The data. If not None, then the path will not be loaded.
        '''
        if path is not None:
            path = pathlib.Path(path).resolve()
            if path.is_dir():
                path /= self.FILENAME
        self.path = path

        self.data = {} if data is None else data

        if self.path and data is None:
            self.load()

    def load(self, path=None):
        '''
        Load data from a YAML file.

        Args:
            path:
                The filepath. If None, the current path will be used.
        '''
        path = self.path if path is None else pathlib.Path(path).resolve()
        if path is None:
            raise ValueError('no path set')
        if self.path is None:
            self.path = path
        with path.open('rb') as handle:
            self.data = yaml.safe_load(handle)

    def save(self, path=None):
        '''
        Save data to a YAML file.

        Args:
            path:
                The filepath. If None, the current path will be used.
        '''
        # Attempt to write to a temporary file first to ensure that the metadata
        # file is not destroyed if the data cannot be converted to YAML.
        with tempfile.TemporaryDirectory() as tmp_dir:
            tmp_dir = pathlib.Path(tmp_dir)
            tmp_path = tmp_dir / self.FILENAME
            with tmp_path.open('w', encoding='utf-8') as handle:
                yaml.dump(self.data, handle)
            path = self.path if path is None else pathlib.Path(path).resolve()
            path.parent.mkdir(parents=True, exist_ok=True)
            LOGGER.debug('Saving data to %s', path)
            shutil.move(tmp_path, path)

    @staticmethod
    def non_negative_int(value):
        '''
        Check that the given value is a non-negative integer.

        Args:
            value:
                The value to check.

        Returns:
            The value as an int.

        Raises:
            ValueError:
                The value could not be converted to a non-negative int.
        '''
        value = int(value)
        if value < 0:
            raise ValueError('negative value')
        return value

    def _get_experiment_id(self):
        return self.data.get(self.FIELD_EXPERIMENT_ID)

    def _set_experiment_id(self, value):
        self.data[self.FIELD_EXPERIMENT_ID] = self.non_negative_int(value)

    experiment_id = property(
        fget=_get_experiment_id,
        fset=_set_experiment_id,
        doc='The experiment ID'
    )

    def _get_name(self):
        return self.data.get(self.FIELD_NAME)

    def _set_name(self, value):
        value = str(value)
        if not value:
            raise ValueError('empty name')
        self.data[self.FIELD_NAME] = value

    name = property(
        fget=_get_name,
        fset=_set_name,
        doc='The name'
    )

    def _get_creation_time(self):
        return self.data.get(self.FIELD_CREATION_TIME)

    def _set_creation_time(self, value):
        self.data[self.FIELD_CREATION_TIME] = self.non_negative_int(value)

    creation_time = property(
        fget=_get_creation_time,
        fset=_set_creation_time,
        doc='The creation time'
    )

    def _get_last_update_time(self):
        return self.data.get(self.FIELD_LAST_UPDATE_TIME)

    def _set_last_update_time(self, value):
        self.data[self.FIELD_LAST_UPDATE_TIME] = self.non_negative_int(value)

    last_update_time = property(
        fget=_get_last_update_time,
        fset=_set_last_update_time,
        doc='The last update time'
    )

    def _get_lifecycle_stage(self):
        return self.data.get(self.FIELD_LIFECYCLE_STAGE)

    def _set_lifecycle_stage(self, value):
        value = str(value)
        if not value:
            raise ValueError('empty lifecycle_stage')
        self.data[self.FIELD_LIFECYCLE_STAGE] = value

    lifecycle_stage = property(
        fget=_get_lifecycle_stage,
        fset=_set_lifecycle_stage,
        doc='The lifecycle stage'
    )

    @property
    def is_experiment(self):
        '''
        True if the current metadata appears to be for an experiment and not a
        run.
        '''
        return self.FIELD_ARTIFACT_LOCATION in self.data

    @property
    def _artifact_uri_field(self):
        '''
        Get the artifact URI field or equivalent depending on the type of file.
        '''
        return self.FIELD_ARTIFACT_LOCATION \
            if self.FIELD_ARTIFACT_LOCATION in self.data \
            else self.FIELD_ARTIFACT_URI

    def _get_artifact_uri(self):
        return URITransformer(self.data.get(self._artifact_uri_field))

    def _set_artifact_uri(self, value):
        if not isinstance(value, URITransformer):
            value = URITransformer(value)
        if not value.parts[2]:
            raise ValueError('empty artifact path')
        self.data[self._artifact_uri_field] = str(value)

    artifact_uri = property(
        fget=_get_artifact_uri,
        fset=_set_artifact_uri,
        doc='The artifact URI or equivalent'
    )

    def find_meta(self, path):
        '''
        Find all MLflow metadata files in the given directory.

        Args:
            path:
                The directory path.

        Returns:
            A generator over the metadata file paths as pathlib.Path objects.
        '''
        path = pathlib.Path(path).resolve()
        for root, _dirs, files in os.walk(path):
            if self.FILENAME in files:
                yield self.__class__(path=root)
